
/*
 *  shm_msgserver.c
 *  
 *  Illustrates memory mapping and persistency, with POSIX objects.
 *
 *  This process produces a message leaving it in a shared segment.
 *  The segment is mapped in a persistent object meant to be subsequently
 *  opened by a shared memory "client".
 *
 *  Created by Mij <mij@bitchx.it> on 27/08/05.
 *  Original source file available at http://mij.oltrelinux.com/devel/unixprg/
 *
 *  Note from nieminen@jyu.fi: This version of Mij's public domain
 *  example is adapted for our spring 2015 and later 2016 course on OS
 *  fundamentals. Students, please see the original site for more
 *  information!
 *
 */

#define _XOPEN_SOURCE 700 /* Single UNIX Specification, Version x */
/* (See POSIX 2008.1 rationale chapter B.2.2 ) */

#include <stdio.h>

/* shm_* stuff, and mmap() */
#include <sys/mman.h>
#include <sys/types.h>

/* exit() etc */
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

/* for random() stuff and small waiting for the race to be evident. */
#include <stdlib.h>
#include <time.h>


/* POSIX IPC object name [system dependent] - see
 * http://http://mij.oltrelinux.com/devel/unixprg/#ipc__posix_objects
 *
 * ... On Linux, you may observe (and read/write:)) /dev/shm/foo1423
 * after creation. A realistic file should have proper access rights
 * (only for the user who owns the memory!!! or *at most* group),
 * and a name that is unlikely to clash with others in the system
 * (combination of username and application name perhaps?
 * random/nondeterministic names can't be used because the idea is to
 * create the name in all the applications that need the shared
 * segment; graceful handling of pre-existing file should be
 * implemented - the user may always start another instance of a
 * server!).
 */
#define SHMOBJ_PATH         "/foo1423"

/* maximum length of the content of the message */
#define MAX_MSG_LENGTH      50

/* how many types of messages we recognize (fantasy) */
#define TYPES               8

/* message structure for messages in the shared segment */
struct msg_s {
    int type;
    char content[MAX_MSG_LENGTH];
};


int main(int argc, char *argv[]) {
    int shmfd;
    int shared_seg_size = (1 * sizeof(struct msg_s));   /* want shared segment capable of storing 1 message */
    struct msg_s *shared_msg;      /* the shared segment, and head of the messages list */
    
    /* Forcible cleanup in case we've had an unclean shutdown */
    /* Note: In a real application, termination signals should be
       caught by a signal handler for cleanup! (And that won't protect
       against a kill -9 from the user! As a user, prefer the TERM
       signal over KILL.)*/
    if (argc > 1) goto cleanup;

    /* creating the shared memory object    --  shm_open()  */
    shmfd = shm_open(SHMOBJ_PATH, O_CREAT | O_EXCL | O_RDWR, S_IRWXU);

    if (shmfd < 0) {
        perror("In shm_open()");
        exit(1);
    }
    fprintf(stderr, "Created shared memory object %s\n", SHMOBJ_PATH);
    
    /* adjusting mapped file size (make room for the whole segment to map)      --  ftruncate() */
    ftruncate(shmfd, shared_seg_size);

    /* requesting the shared segment    --  mmap() */
    shared_msg = (struct msg_s *)mmap(NULL, shared_seg_size, PROT_READ | PROT_WRITE, MAP_SHARED, shmfd, 0);
    if (shared_msg == NULL) {
        perror("In mmap()");
        exit(1);
    }
    fprintf(stderr, "Shared memory segment allocated correctly (%d bytes).\n", shared_seg_size);

    
    srandom(time(NULL));
    /* producing a message on the shared segment */
    shared_msg->type = random() % TYPES;
    snprintf(shared_msg->content, MAX_MSG_LENGTH, "My message, type %d, num %ld", shared_msg->type, random());

    sleep(30); /* adaptation (nieminen 2015): Leave the server "up"
                * for a period, then clean up and "shutdown".
                */

    /* [uncomment if you wish] requesting the removal of the shm object     --  shm_unlink() */

 cleanup:

    if (shm_unlink(SHMOBJ_PATH) != 0) {
        perror("In shm_unlink()");
        exit(1);
    }

    return 0;
}


